import re

from typing_extensions import override

import clypi.parsers as cp
from clypi import Command, arg, cprint


class SlackChannel(cp.Str):
    @override
    def __call__(self, raw: str | list[str], /) -> str:
        parsed = super().__call__(raw)
        if not re.match(r"#[a-z0-9-]+", parsed):
            raise ValueError("Invalid slack channel")
        return parsed


class SlackChannelId(cp.Int):
    @override
    def __call__(self, raw: str | list[str], /) -> int:
        parsed = super().__call__(raw)
        if parsed < 1_000_000 or parsed > 9_999_999:
            raise ValueError(f"Invalid Slack channel {parsed}, it must be 8 digits")
        return parsed


class Main(Command):
    """An example of how useful custom parsers can be"""

    slack: str | int | None = arg(
        help="The Slack channel to send notifications to",
        prompt="What Slack channel should we send notifications to?",
        parser=SlackChannel() | SlackChannelId() | cp.NoneParser(),
    )

    @override
    async def run(self):
        cprint(f"Slack: {self.slack} ({type(self.slack)})", fg="blue")
        cprint("Try using a valid or invalid slack channel", fg="cyan")


if __name__ == "__main__":
    main: Main = Main.parse()
    main.start()



import typing as t
from dataclasses import dataclass, field

from clypi._colors import ColorType, style


class ClypiException(Exception):
    pass


class ClypiExceptionGroup(ExceptionGroup):
    pass


class MaxAttemptsException(ClypiException):
    pass


class AbortException(ClypiException):
    pass


@dataclass
class TreeExcNode:
    exc: BaseException
    children: list[t.Self] = field(default_factory=list)


def _build_exc_tree(err: BaseException) -> TreeExcNode:
    root = TreeExcNode(err)

    # Add __cause__ levels
    if err.__cause__ is not None:
        root.children.append(_build_exc_tree(err.__cause__))

    # Add exception group levels
    if isinstance(err, ExceptionGroup):
        for sub_exc in err.exceptions:
            root.children.append(_build_exc_tree(sub_exc))

    return root


def format_traceback(err: BaseException, color: ColorType | None = "red") -> list[str]:
    def _format_exc(e: BaseException, indent: int):
        msg, *_ = e.args
        icon = "  " * (indent - 1) + " ↳ " if indent != 0 else ""
        return style(f"{icon}{str(msg)}", fg=color)

    def _print_level(tree: TreeExcNode, indent: int) -> list[str]:
        ret: list[str] = []
        for child_node in tree.children:
            ret.append(_format_exc(child_node.exc, indent))
            ret.extend(_print_level(child_node, indent=indent + 1))
        return ret

    tree = _build_exc_tree(err)
    lines = [_format_exc(tree.exc, indent=0)]
    lines.extend(_print_level(tree, indent=1))
    return lines


def print_traceback(err: BaseException) -> None:
    for line in format_traceback(err):
        print(line)


def flatten_exc(err: Exception) -> list[Exception]:
    if not isinstance(err, ExceptionGroup):
        return [err]

    result: list[Exception] = []
    for sub_err in err.exceptions:
        result.extend(flatten_exc(sub_err))
    return result



import typing as t
from dataclasses import dataclass

import pytest

from clypi import Command, arg
from tests.cli_parse_test import parametrize
from tests.prompt_test import replace_stdin


@dataclass
class CustomType:
    foo: str = "bar"


def parse_custom(raw: str | list[str]) -> CustomType:
    return CustomType(foo=str(raw))


class Run(Command):
    """
    Runs all files
    """

    verbose: bool = arg(inherited=True)
    env: str = arg(inherited=True)
    env_prompt: str = arg(inherited=True)
    custom: CustomType = arg(inherited=True)


class Main(Command):
    subcommand: Run | None = None
    verbose: bool = arg(False, short="v", help="Whether to show more output")
    env: t.Literal["qa", "prod"] = arg(help="The environment to use")
    env_prompt: t.Literal["qa", "prod"] = arg(
        help="The environment to use",
        prompt="What environment should we use?",
    )
    custom: CustomType = arg(default=CustomType(), parser=parse_custom)


@parametrize(
    "args,expected,fails,stdin",
    [
        ([], {}, True, ""),
        (["-v"], {}, True, ""),
        (["-v", "--env", "qa"], {}, True, ""),
        (
            ["-v", "--env", "qa"],
            {
                "verbose": True,
                "env": "qa",
                "env_prompt": "qa",
            },
            False,
            "qa\n",
        ),
        (
            ["-v", "--env", "qa", "--env-prompt", "qa"],
            {
                "verbose": True,
                "env": "qa",
                "env_prompt": "qa",
            },
            False,
            "",
        ),
        (
            ["--env", "qa", "-v", "run"],
            {
                "verbose": True,
                "env": "qa",
                "env_prompt": "qa",
                "run": {
                    "verbose": True,
                    "env": "qa",
                    "env_prompt": "qa",
                },
            },
            False,
            "qa\n",
        ),
        (
            ["--custom", "baz", "run", "--env", "qa", "-v"],
            {
                "verbose": True,
                "env": "qa",
                "env_prompt": "qa",
                "run": {
                    "verbose": True,
                    "env": "qa",
                    "env_prompt": "qa",
                    "custom": CustomType("baz"),
                },
                "custom": CustomType("baz"),
            },
            False,
            "qa\n",
        ),
        (
            ["--env", "qa", "run", "-v", "--env-prompt", "qa"],
            {
                "verbose": True,
                "env": "qa",
                "env_prompt": "qa",
                "run": {
                    "verbose": True,
                    "env": "qa",
                    "env_prompt": "qa",
                },
            },
            False,
            "",
        ),
        (
            ["--env", "qa", "--env-prompt", "qa", "run", "-v"],
            {
                "verbose": True,
                "env": "qa",
                "env_prompt": "qa",
                "run": {
                    "verbose": True,
                    "env": "qa",
                    "env_prompt": "qa",
                },
            },
            False,
            "",
        ),
        (
            ["--env", "qa", "run", "-v"],
            {
                "verbose": True,
                "env": "qa",
                "env_prompt": "qa",
                "run": {
                    "verbose": True,
                    "env": "qa",
                    "env_prompt": "qa",
                },
            },
            False,
            "qa\n",
        ),
        (
            ["run", "--env", "qa", "-v"],
            {
                "verbose": True,
                "env": "qa",
                "env_prompt": "qa",
                "run": {
                    "verbose": True,
                    "env": "qa",
                    "env_prompt": "qa",
                },
            },
            False,
            "qa\n",
        ),
        (["run", "-v"], {}, True, ""),
        (
            ["run", "--env", "qa", "-v", "--custom", "baz"],
            {
                "verbose": True,
                "env": "qa",
                "env_prompt": "qa",
                "run": {
                    "verbose": True,
                    "env": "qa",
                    "env_prompt": "qa",
                    "custom": CustomType("baz"),
                },
                "custom": CustomType("baz"),
            },
            False,
            "qa\n",
        ),
    ],
)
def test_parse_inherited(
    args: list[str],
    expected: dict[str, t.Any],
    fails: bool,
    stdin: str | list[str],
):
    if fails:
        with pytest.raises(Exception):
            _ = Main.parse(args)
        return

    # Check command
    with replace_stdin(stdin):
        main = Main.parse(args)

    assert main is not None
    for k, v in expected.items():
        if k == "run":
            continue
        lc_v = getattr(main, k)
        assert lc_v == v, f"{k} should be {v} but got {lc_v}"

    # Check subcommand
    if "run" in expected:
        assert main.subcommand is not None
        assert isinstance(main.subcommand, Run)
        for k, v in expected["run"].items():
            lc_v = getattr(main, k)
            assert lc_v == v, f"run.{k} should be {v} but got {lc_v}"



from textwrap import dedent

from clypi import format_traceback
from clypi._exceptions import ClypiExceptionGroup


def test_basic_traceback():
    root_cause = RuntimeError("Actual root cause")

    context = ValueError("Some context")
    context.__cause__ = root_cause

    err = ValueError("User facing error")
    err.__cause__ = context

    result = format_traceback(err, color=None)
    result_str = "\n".join(result)
    assert (
        result_str
        == dedent(
            """
            User facing error
             ↳ Some context
               ↳ Actual root cause
            """
        ).strip()
    )


def test_traceback_exc_group():
    context1 = ValueError("Failed to parse 'foo' as int()")
    context2 = ValueError("Failed to parse 'foo' as None")
    err = ClypiExceptionGroup("Failed to parse as int|none", [context1, context2])

    result = format_traceback(err, color=None)
    result_str = "\n".join(result)
    assert (
        result_str
        == dedent(
            """
            Failed to parse as int|none
             ↳ Failed to parse 'foo' as int()
             ↳ Failed to parse 'foo' as None
            """
        ).strip()
    )


def test_traceback_complex():
    root1 = ValueError("Invalid int literal for base 10 'foo'")
    context1 = ValueError("Failed to parse 'foo' as int()")
    context1.__cause__ = root1

    root2_1 = ValueError("Text 'foo' does not match 'none'")
    root2_2 = ValueError("Text 'foo' is not an empty list")
    context2 = ClypiExceptionGroup("Failed to parse 'foo' as None", [root2_1, root2_2])

    err = ClypiExceptionGroup("Failed to parse as int|none", [context1, context2])

    result = format_traceback(err, color=None)
    result_str = "\n".join(result)
    assert (
        result_str
        == dedent(
            """
            Failed to parse as int|none
             ↳ Failed to parse 'foo' as int()
               ↳ Invalid int literal for base 10 'foo'
             ↳ Failed to parse 'foo' as None
               ↳ Text 'foo' does not match 'none'
               ↳ Text 'foo' is not an empty list
            """
        ).strip()
    )
